<?php

namespace InSegment\ApiCore\Api\V_2_0;

use Closure;

use InSegment\ApiCore\Api\V_2_0\DTOUtil;
use InSegment\ApiCore\Api\V_2_0\MappingTypes\Attribute;
use InSegment\ApiCore\Models\SourceAttributeDownstream;

class RuleSetLoader
{
    /**
     * DTODefs instance
     *
     * @var \InSegment\ApiCore\Api\V_2_0\DTOUtil 
     */
    protected $dtoDefs;
    
    /**
     * Raw loaded rule sets
     *
     * @var array
     */
    protected $loadedRuleSets = [];
    
    /**
     * Cache of initialised rule sets
     *
     * @var array
     */
    protected $parsedRuleSets = [];
    
    /**
     * Constructor
     * 
     * @param \InSegment\ApiCore\Api\V_2_0\DTOUtil $dtoDefs
     */
    public function __construct(DTOUtil $dtoDefs)
    {
        $this->dtoDefs = $dtoDefs;
    }
    
    /**
     * Load rules from file using name
     * 
     * @param string $name
     * @return $this
     * @throws \Exception
     */
    public function loadRules(string $name)
    {
        if (!isset($this->loadedRuleSets[$name])) {
            $this->loadedRuleSets[$name] = $this->dtoDefs->loadRulesFile($name);
            
            if (!is_array($this->loadedRuleSets[$name])) {
                throw new \Exception("Failed to load rule set: {$name}");
            }
        }
        
        return $this;
    }
    
    /**
     * Load custom rules set into virtual rule set name
     * 
     * @param \InSegment\ApiCore\Api\V_2_0\CustomRuleSet $customRuleSet
     * @return null
     */
    public function loadCustomRuleSet(CustomRuleSet $customRuleSet)
    {
        $name = $customRuleSet->getName();
        $this->loadedRuleSets[$name] = $customRuleSet->rules;
    }
    
    /**
     * Get raw rule set (non-initialised for iterator)
     * 
     * @param string $name
     * @return array
     */
    public function getRuleSet(string $name)
    {
        $this->loadRules($name);
        
        return $this->loadedRuleSets[$name];
    }
    
    /**
     * Get rules for iterator (initialised)
     * 
     * @param string $name
     * @return array
     */
    public function getRulesForIterator($name)
    {
        if (!isset($this->parsedRuleSets[$name])) {
            $this->parsedRuleSets[$name] = $this->initialiseRulesForIterator($name);
        }
        
        return $this->parsedRuleSets[$name];
    }
    
    /**
     * Determine the type of keyBy rule and extract a Clousre, if possible
     * 
     * @param mixed $keyBy
     * @return array [
     *     'type' => 'keys'|'closure'|'method',
     *     'value' => array|\Closure|callable
     * ]
     */
    public function processKeyByRule($keyBy)
    {
        if (is_array($keyBy)) {
            return ['type' => 'keys', 'value' => $keyBy];
        } else if ($keyBy instanceof Closure) {
            return ['type' => 'closure', 'value' => $keyBy];
        } else if (is_callable($keyBy)) {
            return ['type' => 'closure', 'value' => Closure::fromCallable($keyBy)];
        } else if (strpos($keyBy, '@') !== false) {
            list($class, $method) = explode('@', $keyBy);
            
            if ($class) {
                return ['type' => 'closure', 'value' => Closure::fromCallable([$class, $method])];
            } else {
                return ['type' => 'method', 'value' => $method];
            }
        } else if (is_string($keyBy)) {
            return [
                'type' => 'keys',
                'value' => [$keyBy]
            ];
        }
    }
    
    /**
     * Prepare rule set for file name
     * 
     * @param string $name
     * @return array [
     *     'classToAttribute' => array [
     *         string $class => [
     *              string $dotFormat => array [
     *                  'type' => 'mapperType',
     *                  ...
     *              ],
     *              ...
     *         ], ...
     *     ],
     *     'mapperTypes' => [
     *         string $mapperType => string $handlerClass,
     *         ...
     *     ]
     * ]
     * @throws \Exception
     */
    protected function initialiseRulesForIterator(string $name): array
    {
        $result = [];
        $mapperTypes = [];
        foreach ($this->getRuleSet($name) as $class => $mappings) {
            $notArray = !is_array($mappings);
            $attributeSet = [];
            foreach (array_wrap($mappings) as $dataAddress => $mapping) {
                if (!isset($mapping)) {
                    continue;
                }

                if (is_array($mapping)) {
                    if (!isset($mapping['type'])) {
                        throw new \Exception("Malformed rule '{$dataAddress}' of '{$class}': 'type' field is required");
                    }

                    $modifiedMapping = $mapping;
                } else {
                    $modifiedMapping = ['type' => 'attribute', 'attribute' => $mapping];
                }
                
                if ($modifiedMapping['type'] === 'attribute' && str_contains($modifiedMapping['attribute'], ['.', ':', '()'])) {
                    $modifiedMapping['mapped'] = Attribute::characterizeAttributeMapping($modifiedMapping['attribute']);
                }

                if (!isset($mapperTypes[$modifiedMapping['type']])) {
                    $mapperTypes[$modifiedMapping['type']] = $this->dtoDefs->classifyMapperType($modifiedMapping['type']);
                }

                $modifiedMapping['extra'] = array_flip($modifiedMapping['extra'] ?? []);
                $modifiedMapping['dataAddress'] = $notArray ? '-' : $dataAddress;
                
                if (isset($modifiedMapping['keyBy'])) {
                    $modifiedMapping['keyBy'] = $this->processKeyByRule($modifiedMapping['keyBy']);
                }
                
                $attributeSet[] = $modifiedMapping;
            }
            
            $result[$class] = SourceAttributeDownstream::prepareAttributeSet($attributeSet);
        }
        
        return ['classToAttribute' => $result, 'mapperTypes' => $mapperTypes];
    }
    
}
